from datetime import datetime
from typing import Any
from typing import Dict

from robinhood.lib.helpers import parse_datetime_with_microsecond


class OptionPosition(object):

    def __init__(
        self,
        position_id: str,
        position_type: str,
        position_url: str,
        account_url: str,
        option_instrument_url: str,
        chain_id: str,
        chain_symbol: str,
        average_price: float,
        quantity: float,
        trade_value_multiplier: float,
        intraday_average_open_price: float,
        intraday_quantity: float,
        pending_assignment_quantity: float,
        pending_exercise_quantity: float,
        pending_expiration_quantity: float,
        pending_expired_quantity: float,
        pending_buy_quantity: float,
        pending_sell_quantity: float,
        created_at: datetime,
        updated_at: datetime,
    ) -> None:
        self._position_id = position_id
        self._position_type = position_type
        self._position_url = position_url
        self._account_url = account_url
        self._option_instrument_url = option_instrument_url
        self._chain_id = chain_id
        self._chain_symbol = chain_symbol
        self._average_price = average_price
        self._quantity = quantity
        self._trade_value_multiplier = trade_value_multiplier
        self._intraday_average_open_price = intraday_average_open_price
        self._intraday_quantity = intraday_quantity
        self._pending_assignment_quantity = pending_assignment_quantity
        self._pending_exercise_quantity = pending_exercise_quantity
        self._pending_expiration_quantity = pending_expiration_quantity
        self._pending_expired_quantity = pending_expired_quantity
        self._pending_buy_quantity = pending_buy_quantity
        self._pending_sell_quantity = pending_sell_quantity
        self._created_at = created_at
        self._updated_at = updated_at

    @property
    def position_id(self) -> str:
        return self._position_id

    @property
    def position_type(self) -> str:
        return self._position_type

    @property
    def option_instrument_url(self) -> str:
        return self._option_instrument_url

    @property
    def chain_symbol(self) -> str:
        return self._chain_symbol

    @property
    def average_price(self) -> float:
        return abs(self._average_price)

    @property
    def avg_unit_price(self) -> float:
        return abs(self._average_price) / self._trade_value_multiplier

    @property
    def active_quantity(self) -> float:
        return (
            self._quantity
            - self._pending_expiration_quantity
            - self._pending_assignment_quantity
        )

    @property
    def quantity(self) -> float:
        return self._quantity

    @property
    def pending_expiration_quantity(self) -> float:
        return self._pending_expiration_quantity

    @property
    def pending_assignment_quantity(self) -> float:
        return self._pending_assignment_quantity


def option_position_from_data(data: Dict[str, Any]) -> OptionPosition:
    return OptionPosition(
        position_id=data['id'],
        position_type=data['type'],
        position_url=data['url'],
        account_url=data['account'],
        option_instrument_url=data['option'],
        chain_id=data['chain_id'],
        chain_symbol=data['chain_symbol'],
        average_price=float(data['average_price']),
        quantity=float(data['quantity']),
        trade_value_multiplier=float(data['trade_value_multiplier']),
        intraday_average_open_price=float(data['intraday_average_open_price']),
        intraday_quantity=float(data['intraday_quantity']),
        pending_assignment_quantity=float(data['pending_assignment_quantity']),
        pending_exercise_quantity=float(data['pending_exercise_quantity']),
        pending_expiration_quantity=float(data['pending_expiration_quantity']),
        pending_expired_quantity=float(data['pending_expired_quantity']),
        pending_buy_quantity=float(data['pending_buy_quantity']),
        pending_sell_quantity=float(data['pending_sell_quantity']),
        created_at=parse_datetime_with_microsecond(data['created_at']),
        updated_at=parse_datetime_with_microsecond(data['updated_at']),
    )



from collections import defaultdict
from collections import OrderedDict
from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

from asciitree import LeftAligned
from terminaltables import AsciiTable

from data.watchlist import ALL_REVERSED
from report.symbolhistory.symbolhistorycontext import SymbolHistoryContext
from report.symbolhistory.symbolhistoryevent import symbol_history_event_from_option_event
from report.symbolhistory.symbolhistoryevent import symbol_history_event_from_option_order
from report.symbolhistory.symbolhistoryevent import symbol_history_event_from_stock_order
from report.symbolhistory.symbolhistoryevent import SymbolHistoryEvent
from robinhood.lib.account.account import account
from robinhood.lib.account.optionevent.optionevent import OptionEvent
from robinhood.lib.account.optionorder.optionorder import OptionOrder
from robinhood.lib.account.stockorder.stockorder import StockOrder
from robinhood.lib.helpers import format_optional_greek
from robinhood.lib.helpers import format_optional_price
from robinhood.lib.helpers import format_percentage
from robinhood.lib.helpers import format_price
from robinhood.lib.helpers import format_quantity
from robinhood.lib.marketplace.equity import Equity
from robinhood.lib.marketplace.marketplace import marketplace
from robinhood.lib.robinhoodclient import rhc


class SymbolHistory(object):

    def __init__(
        self,
        stock_orders: List[StockOrder],
        option_orders: List[OptionOrder],
        option_events: List[OptionEvent],
    ) -> None:
        self._events_by_symbol = self._load_events_by_symbol(
            stock_orders=stock_orders,
            option_orders=option_orders,
            option_events=option_events,
        )

    def get_events(self, symbol: str) -> List[SymbolHistoryEvent]:
        return self._events_by_symbol[symbol]

    def show_current_positions(self) -> None:
        active_symbol = account.active_symbols
        equaities = marketplace.get_equities(symbols=active_symbol)

        with ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(
                    self.get_current_position_report,
                    symbol=symbol,
                    equity=equaities[symbol],
                )
                for symbol in active_symbol
            ]

        reports_by_symbol: Dict[str, str] = {}
        for future in as_completed(futures):
            symbol, report = future.result()
            reports_by_symbol.update({symbol: report})

        for symbol in sorted(reports_by_symbol.keys()):
            print(reports_by_symbol[symbol])
            print()

    def get_current_position_report(self, symbol: str, equity: Equity) -> Tuple[str, str]:
        events = self.get_events(symbol=symbol)
        if len(events) == 0:
            report = f'No event for {symbol}...'
            return (symbol, report)

        event = events[-1]

        # Stock
        symbol_with_price = '{symbol} {last_price} {dod_change}'.format(
            symbol=symbol,
            last_price=format_price(equity.last_price),
            dod_change=format_percentage(equity.dod_percentage),
        )
        symbol_data: OrderedDict[str, Any] = OrderedDict()

        stock_key, stock_details = self._get_current_stock_holding(
            event=event,
            last_price=equity.last_price,
        )
        symbol_data[stock_key] = stock_details

        # Option Buy and Sell
        for buy_or_sell in ['Buy', 'Sell']:
            for option_key, option_details in self._get_current_option_buy_or_sell_holding(
                event=event,
                last_price=equity.last_price,
                buy_or_sell=buy_or_sell,
            ):
                symbol_data[option_key] = option_details

        data = {symbol_with_price: symbol_data}

        tree = LeftAligned()
        return (symbol, tree(data))

    def _get_current_stock_holding(
        self,
        event: SymbolHistoryEvent,
        last_price: float,
    ) -> 'Tuple[str, OrderedDict[str, Any]]':
        stock_quantity = event.context.stock_holding.quantity
        stock_avg_price = event.context.stock_holding.avg_unit_price
        stock_total_value = stock_quantity * stock_avg_price
        if stock_quantity == 0:
            stock_holding = 'none'
        else:
            stock_holding = '{quantity} x {avg_price} = {total_value}'.format(
                quantity=stock_quantity,
                avg_price=format_price(stock_avg_price),
                total_value=format_price(stock_total_value),
            )
        stock_key = f'Stock Holding: {stock_holding}'
        stock_details: OrderedDict[str, Any] = OrderedDict()

        if stock_quantity > 0:
            current_value = last_price * stock_quantity
            stock_details['Current Value: {value} ({change})'.format(
                value=format_price(current_value),
                change=format_percentage(current_value / stock_total_value - 1),
            )] = {}

        return (stock_key, stock_details)

    def _get_current_option_buy_or_sell_holding(
        self,
        event: SymbolHistoryEvent,
        last_price: float,
        buy_or_sell: str,
    ) -> 'Iterator[Tuple[str, OrderedDict[str, Any]]]':
        if buy_or_sell == 'Buy':
            holding_dict = event.context.option_buy_holding
        else:
            holding_dict = event.context.option_sell_holding

        for _, item in holding_dict.items():
            total_base_price = item.avg_unit_price * item.quantity * 100
            if item.option_instrument.instrument_type == 'call':
                if last_price > item.option_instrument.strike_price:
                    diff = last_price - item.option_instrument.strike_price
                    above_or_below = 'above'
                    breached = 'BREACHED'
                else:
                    diff = item.option_instrument.strike_price - last_price
                    above_or_below = 'below'
                    breached = 'NOT BREACHED'
            else:
                assert item.option_instrument.instrument_type == 'put'
                if last_price < item.option_instrument.strike_price:
                    diff = item.option_instrument.strike_price - last_price
                    above_or_below = 'below'
                    breached = 'BREACHED'
                else:
                    diff = last_price - item.option_instrument.strike_price
                    above_or_below = 'above'
                    breached = 'NOT BREACHED'
            breached_key = '{breached} - {above_or_below} strike price {diff_dollar} or {diff_percentage}'.format(
                above_or_below=above_or_below,
                diff_dollar=format_price(diff),
                diff_percentage=format_percentage(diff / last_price),
                breached=breached,
            )

            option_sell_key = (
                '{buy_or_sell} {symbol} ({days_to_expire}): {quantity} x {avg_price} = {base_price}'.format(
                    buy_or_sell=buy_or_sell,
                    symbol=item.option_instrument.symbol,
                    strike_price=format_price(item.option_instrument.strike_price),
                    exp_date=item.option_instrument.exp_date,
                    days_to_expire=format_quantity(item.option_instrument.days_to_expire),
                    quantity=item.quantity,
                    avg_price=format_price(item.avg_unit_price),
                    base_price=format_price(total_base_price),
                )
            )
            market_data = item.option_instrument.market_data
            if market_data is None:
                current_price = '-'
                theta = '-'
                delta = '-'
                gamma = '-'
                vega = '-'
                rho = '-'
            else:
                current_price = format_optional_price(market_data.medium_fill_rate_sell_price)
                theta = format_optional_greek(market_data.theta)
                delta = format_optional_greek(market_data.delta)
                gamma = format_optional_greek(market_data.gamma)
                vega = format_optional_greek(market_data.vega)
                rho = format_optional_greek(market_data.rho)

            current_price_key = f'Current Price: {current_price}'
            greek_key = f'Theta: {theta} Delta: {delta} Gamma: {gamma} Vega: {vega} Rho: {rho}'

            option_sell_details: OrderedDict[str, Any] = OrderedDict()
            option_sell_details[current_price_key] = {}
            option_sell_details[breached_key] = {}
            option_sell_details[greek_key] = {}

            yield (option_sell_key, option_sell_details)

    def show(self, symbol: str, last_n_events: Optional[int] = 10) -> None:
        data = self._gen_ascii_tbl_data(symbol=symbol)
        if last_n_events is None:
            applicable_data = data
        else:
            data_no_head = data[1:]
            applicable_data = [data[0]] + data_no_head[-last_n_events - 1:]
        t = AsciiTable(applicable_data)
        t.inner_column_border = True
        t.inner_footing_row_border = False
        t.inner_heading_row_border = True
        t.inner_row_border = False
        t.outer_border = False
        t.justify_columns = {
            0: 'center',
            1: 'center',
            2: 'center',
            3: 'right',
            4: 'right',
            5: 'right',
            6: 'right',
            7: 'right',
        }
        print(t.table)

    def show_holdings_by_events(self, symbol: str) -> None:
        events = self.get_events(symbol=symbol)
        for event in events:
            print(event.event_name)
            quantity = event.context.stock_holding.quantity
            avg_price = event.context.stock_holding.avg_unit_price
            print('>>> Stock Holding: {quantity} x {avg_price} = {total_value}'.format(
                quantity=format_quantity(quantity),
                avg_price=format_price(avg_price),
                total_value=format_price(quantity * avg_price),
            ))

            for _, item in event.context.option_buy_holding.items():
                print('>>> Option Buy Holding: {quantity} x {avg_price} = {total_value}'.format(
                    quantity=format_quantity(item.quantity),
                    avg_price=format_price(item.avg_unit_price),
                    total_value=format_price(item.quantity * item.avg_unit_price),
                ))

            for _, item in event.context.option_sell_holding.items():
                print('>>> Option Sell Holding: {quantity} x {avg_price} = {total_value}'.format(
                    quantity=format_quantity(item.quantity),
                    avg_price=format_price(item.avg_unit_price),
                    total_value=format_price(item.quantity * item.avg_unit_price),
                ))

            print()

    def _gen_ascii_tbl_data(self, symbol: str) -> List[List[str]]:
        data = [[
            'Date',
            'Event',
            'Symbol',
            'Quantity',
            'Price',
            '# of Shares',
            'Avg Price',
            'Cumulative Profit',
        ]]

        for event in self._events_by_symbol[symbol]:
            # could be option legs from vertical
            if event.context is None:
                continue

            data.append(
                [
                    event.formatted_ts,
                    event.event_name,
                    event.symbol,
                    format_quantity(event.quantity),
                    format_optional_price(event.unit_price),
                    format_quantity(event.context.stock_holding.quantity),
                    format_price(event.context.stock_holding.avg_unit_price),
                    format_price(event.context.profit),
                ],
            )

        return data

    def _load_events_by_symbol(
        self,
        stock_orders: List[StockOrder],
        option_orders: List[OptionOrder],
        option_events: List[OptionEvent],
    ) -> DefaultDict[str, List[SymbolHistoryEvent]]:
        events_by_symbol: DefaultDict[str, List[SymbolHistoryEvent]] = defaultdict(list)

        self._convert_stock_orders_to_symbol_history_events(
            stock_orders=stock_orders,
            events_by_symbol=events_by_symbol,
        )

        self._convert_option_orders_to_symbol_history_events(
            option_orders=option_orders,
            events_by_symbol=events_by_symbol,
        )

        self._convert_option_events_to_symbol_history_events(
            option_events=option_events,
            events_by_symbol=events_by_symbol,
        )

        for symbol, symbol_events in events_by_symbol.items():
            events_by_symbol[symbol] = sorted(symbol_events, key=lambda e: e.event_ts)

        for symbol, symbol_events in events_by_symbol.items():
            context = SymbolHistoryContext(symbol=symbol)
            for event in symbol_events:
                if event.event_name in ('SHORT PUT', 'SHORT CALL'):
                    assert event.unit_price is not None
                    assert event.option_order.num_of_legs == 1
                    leg = event.option_order.legs[0]
                    context.sell_to_open(
                        symbol=event.symbol,
                        quantity=int(event.quantity) * leg.ratio_quantity,
                        price=abs(event.unit_price),
                        option_instrument=leg.option_instrument,
                    )
                elif event.event_name in ('LONG PUT', 'LONG CALL'):
                    assert event.unit_price is not None
                    assert len(event.option_order.opening_legs) == 1
                    context.buy_to_open(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=event.unit_price,
                        option_instrument=event.option_order.legs[0].option_instrument,
                    )
                elif event.event_name in {'ROLLING CALL', 'ROLLING PUT'}:
                    assert event.unit_price is not None
                    closing_option_instrument = event.option_rolling_order.closing_leg.option_instrument
                    unit_price = context.option_sell_holding[closing_option_instrument.symbol].avg_unit_price
                    context.buy_to_close(
                        symbol=closing_option_instrument.symbol,
                        quantity=int(event.quantity),
                        price=unit_price,
                    )

                    leg = event.option_rolling_order.opening_leg
                    context.sell_to_open(
                        symbol=leg.option_instrument.symbol,
                        quantity=int(event.quantity) * leg.ratio_quantity,
                        price=abs(event.unit_price) + unit_price,
                        option_instrument=leg.option_instrument,
                    )
                elif event.event_name == 'SHORT CALL SPREAD':
                    for leg in event.option_order.legs:
                        if leg.side == 'buy':
                            context.buy_to_open(
                                symbol=leg.option_instrument.symbol,
                                quantity=int(leg.quantity),
                                price=leg.price,
                                option_instrument=leg.option_instrument,
                            )
                        else:
                            context.sell_to_open(
                                symbol=leg.option_instrument.symbol,
                                quantity=int(leg.quantity),
                                price=leg.price,
                                option_instrument=leg.option_instrument,
                            )
                elif event.event_name == 'OPTION EXPIRATION':
                    context.option_expire(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                    )
                elif event.event_name == 'BUY TO CLOSE':
                    assert event.unit_price is not None
                    context.buy_to_close(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=event.unit_price,
                    )
                elif event.event_name == 'SELL TO CLOSE':
                    assert event.unit_price is not None
                    context.sell_to_close(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=event.unit_price,
                    )
                elif event.event_name == 'CALL ASSIGNMENT':
                    assert event.unit_price is not None
                    context.call_assignment(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=event.unit_price,
                    )
                elif event.event_name == 'PUT ASSIGNMENT':
                    instrument = event.option_event.option_instrument
                    assert event.unit_price is not None
                    context.put_asignment(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=event.unit_price,
                        strike_price=instrument.strike_price,
                    )
                elif event.event_name == 'OPTION EXERCISE':
                    option_instrument = event.option_event.option_instrument
                    option_avg_price = context.option_buy_holding[event.symbol].avg_unit_price
                    event.unit_price = option_instrument.strike_price + option_avg_price
                    context.call_exercise(
                        symbol=event.symbol,
                        quantity=int(event.quantity),
                        price=option_avg_price,
                        strike_price=option_instrument.strike_price,
                    )
                elif event.event_name in ('LIMIT BUY', 'MARKET BUY'):
                    assert event.unit_price is not None
                    context.limit_or_market_buy_stock(
                        quantity=int(event.quantity),
                        price=event.unit_price,
                    )
                elif event.event_name in ('LIMIT SELL', 'MARKET SELL'):
                    assert event.unit_price is not None
                    context.limit_or_market_sell_stock(
                        quantity=int(event.quantity),
                        price=event.unit_price,
                    )
                elif event.event_name in {
                    'FIG LEAF',
                    'LONG CALL SPREAD',
                    'LONG PUT SPREAD',
                    'SHORT PUT SPREAD',
                    'CUSTOM',
                    'IRON CONDOR',
                }:
                    for leg in event.option_order.legs:
                        quantity = int(event.option_order.processed_quantity) * leg.ratio_quantity
                        if leg.is_buy and leg.is_open:
                            context.buy_to_open(
                                symbol=leg.option_instrument.symbol,
                                quantity=quantity,
                                price=leg.price,
                                option_instrument=leg.option_instrument,
                            )
                        elif leg.is_buy and leg.is_close:
                            context.buy_to_close(
                                symbol=leg.option_instrument.symbol,
                                quantity=quantity,
                                price=leg.price,
                            )
                        elif leg.is_sell and leg.is_open:
                            context.sell_to_open(
                                symbol=leg.option_instrument.symbol,
                                quantity=quantity,
                                price=leg.price,
                                option_instrument=leg.option_instrument,
                            )
                        elif leg.is_sell and leg.is_close:
                            context.sell_to_close(
                                symbol=leg.option_instrument.symbol,
                                quantity=quantity,
                                price=leg.price,
                            )

                elif event.event_name in {
                    'FIG LEAF CLOSE',
                    'LONG CALL SPREAD CLOSE',
                    'LONG CALL CLOSE',
                    'LONG PUT CLOSE',
                    'LONG PUT SPREAD CLOSE',
                    'SHORT PUT SPREAD CLOSE',
                    'CUSTOM CLOSE',
                    'IRON CONDOR CLOSE',
                }:
                    for leg in event.option_order.legs:
                        if leg.side == 'buy':
                            context.buy_to_close(
                                symbol=leg.option_instrument.symbol,
                                quantity=int(event.option_order.processed_quantity) * leg.ratio_quantity,
                                price=leg.price,
                            )
                        else:
                            context.sell_to_close(
                                symbol=leg.option_instrument.symbol,
                                quantity=int(event.option_order.processed_quantity) * leg.ratio_quantity,
                                price=leg.price,
                            )
                else:
                    raise NotImplementedError(event.event_name)

                snapshot = context.take_snapshot()
                event.add_context(symbol_history_context_snapshot=snapshot)

        return events_by_symbol

    def _convert_stock_orders_to_symbol_history_events(
        self,
        stock_orders: List[StockOrder],
        events_by_symbol: DefaultDict[str, List[SymbolHistoryEvent]],
    ) -> None:
        for stock_order in stock_orders:
            if stock_order.instrument_id not in ALL_REVERSED:
                data = rhc.get_url(stock_order.instrument_url)
                for k, v in data.items():
                    print(f'{k}: {v}')

            symbol = ALL_REVERSED[stock_order.instrument_id]
            symbol_history_event = symbol_history_event_from_stock_order(
                symbol=symbol,
                stock_order=stock_order,
            )
            if symbol_history_event is not None:
                events_by_symbol[symbol].append(symbol_history_event)

    def _convert_option_orders_to_symbol_history_events(
        self,
        option_orders: List[OptionOrder],
        events_by_symbol: DefaultDict[str, List[SymbolHistoryEvent]],
    ) -> None:
        for option_order in option_orders:
            symbol = option_order.chain_symbol
            symbol_history_event = symbol_history_event_from_option_order(
                symbol=symbol,
                option_order=option_order,
            )
            if symbol_history_event is not None:
                events_by_symbol[symbol].append(symbol_history_event)

    def _convert_option_events_to_symbol_history_events(
        self,
        option_events: List[OptionEvent],
        events_by_symbol: DefaultDict[str, List[SymbolHistoryEvent]],
    ) -> None:
        for option_event in option_events:
            symbol = option_event.option_instrument.chain_symbol
            symbol_history_event = symbol_history_event_from_option_event(option_event=option_event)
            if symbol_history_event is not None:
                events_by_symbol[symbol].append(symbol_history_event)



from typing import List
from typing import Union

from robinhood.lib.marketplace.equity import Equity
from robinhood.lib.marketplace.optioninstrument import OptionInstrument
from robinhood.lib.marketplace.optionmarketdata import OptionMarketData


class OptionSellCandidate(object):

    def __init__(
        self,
        option_instrument: OptionInstrument,
        market_data: OptionMarketData,
        fund_efficiency: float,
    ) -> None:
        self._option_instrument = option_instrument
        self._market_data = market_data
        self._fund_efficiency = fund_efficiency

    @property
    def symbol(self) -> str:
        return self._option_instrument.symbol

    @property
    def chain_symbol(self) -> str:
        return self._option_instrument.chain_symbol

    @property
    def fund_efficiency(self) -> float:
        return self._fund_efficiency

    def gen_data_row(self, chain_symbol_market_data: Equity) -> List[Union[str, float]]:
        assert self._market_data.chance_of_profit_short is not None
        assert self._market_data.high_fill_rate_sell_price is not None
        assert self._market_data.medium_fill_rate_sell_price is not None
        assert self._market_data.low_fill_rate_sell_price is not None

        row: List[Union[str, float]] = [
            self._option_instrument.symbol,
            self._option_instrument.strike_price,
            chain_symbol_market_data.last_price,
            self._get_percentage_to_strike(
                strike_price=self._option_instrument.strike_price,
                current_price=chain_symbol_market_data.last_price,
            ),
            self._market_data.chance_of_profit_short,
            self._market_data.high_fill_rate_sell_price,
            self._market_data.medium_fill_rate_sell_price,
            self._market_data.low_fill_rate_sell_price,
            self._get_estimated_profit_per_10k(
                fill_price=self._market_data.medium_fill_rate_sell_price,
                strike_price=self._option_instrument.strike_price,
                chance_of_profit=self._market_data.chance_of_profit_short,
            ),
            self._fund_efficiency,
            self._market_data.open_interest,
        ]

        return row

    def _get_percentage_to_strike(self, strike_price: float, current_price: float) -> float:
        return 1 - strike_price / current_price

    def _get_estimated_profit_per_10k(self, fill_price: float, strike_price: float, chance_of_profit: float) -> float:
        return fill_price * chance_of_profit / strike_price * 10000



from datetime import datetime
from typing import Any
from typing import List

from robinhood.lib.account.marginbalance import margin_balance_from_data
from robinhood.lib.account.marginbalance import MarginBalance
from robinhood.lib.account.optionevent.optionevent import option_event_from_data
from robinhood.lib.account.optionevent.optionevent import OptionEvent
from robinhood.lib.account.optionorder.optionorder import option_order_from_data
from robinhood.lib.account.optionorder.optionorder import OptionOrder
from robinhood.lib.account.optionposition.optionposition import option_position_from_data
from robinhood.lib.account.optionposition.optionposition import OptionPosition
from robinhood.lib.account.stockorder.stockorder import stock_order_from_data
from robinhood.lib.account.stockorder.stockorder import StockOrder
from robinhood.lib.account.stockposition.stockposition import stock_position_from_data
from robinhood.lib.account.stockposition.stockposition import StockPosition
from robinhood.lib.helpers import parse_datetime_with_microsecond
from robinhood.lib.robinhoodclient import rhc


class AccountObj(object):

    def __init__(
        self,
        buying_power: float,
        crypto_buying_power: float,
        margin_balance: MarginBalance,
        cash: float,
        cash_held_for_orders: float,
        cash_available_for_withdrawal: float,
        portfolio_url: str,
        positions_url: str,
        url: str,
        user_url: str,
        state: str,
        account_type: str,
        option_level: int,
        locked: bool,
        deactivated: bool,
        permanently_deactivated: bool,
        rhs_account_number: int,
        account_number: str,
        active_subscription_id: str,
        withdrawal_halted: bool,
        sma: int,
        sweep_enabled: bool,
        deposit_halted: bool,
        max_ach_early_access_amount: float,
        only_position_closing_trades: bool,
        sma_held_for_orders: int,
        unsettled_debit: float,
        is_pinnacle_account: bool,
        uncleared_deposits: float,
        unsettled_funds: float,
        created_at: datetime,
        updated_at: datetime,
    ) -> None:
        self._buying_power = buying_power
        self._crypto_buying_power = crypto_buying_power
        self._margin_balance = margin_balance
        self._cash = cash
        self._cash_held_for_orders = cash_held_for_orders
        self._cash_available_for_withdrawal = cash_available_for_withdrawal
        self._portfolio_url = portfolio_url
        self._positions_url = positions_url
        self._url = url
        self._user_url = user_url
        self._state = state
        self._account_type = account_type
        self._option_level = option_level
        self._locked = locked
        self._deactivated = deactivated
        self._permanently_deactivated = permanently_deactivated
        self._rhs_account_number = rhs_account_number
        self._account_number = account_number
        self._active_subscription_id = active_subscription_id
        self._withdrawal_halted = withdrawal_halted
        self._sma = sma
        self._sweep_enabled = sweep_enabled
        self._deposit_halted = deposit_halted
        self._max_ach_early_access_amount = max_ach_early_access_amount
        self._only_position_closing_trades = only_position_closing_trades
        self._sma_held_for_orders = sma_held_for_orders
        self._unsettled_debit = unsettled_debit
        self._is_pinnacle_account = is_pinnacle_account
        self._uncleared_deposits = uncleared_deposits
        self._unsettled_funds = unsettled_funds
        self._created_at = created_at
        self._updated_at = updated_at

    @property
    def cash(self) -> float:
        return self._cash

    @property
    def account_type(self) -> str:
        return self._account_type

    @property
    def margin_balance(self) -> MarginBalance:
        return self._margin_balance

    def get_stock_orders(self) -> List[StockOrder]:
        return rhc.get_objects(
            url='https://api.robinhood.com/orders/',
            callback=stock_order_from_data,
            result_key='results',
        )

    def get_stock_positions(self) -> List[StockPosition]:
        return rhc.get_objects(
            url='https://api.robinhood.com/positions/?nonzero=true',
            callback=stock_position_from_data,
            result_key='results',
        )

    def get_option_orders(self) -> List[OptionOrder]:
        return rhc.get_objects(
            url='https://api.robinhood.com/options/orders/',
            callback=option_order_from_data,
            result_key='results',
        )

    def get_option_positions(self) -> List[OptionPosition]:
        return rhc.get_objects(
            url='https://api.robinhood.com/options/positions/',
            callback=option_position_from_data,
            result_key='results',
        )

    def get_option_events(self) -> List[OptionEvent]:
        return rhc.get_objects(
            url='https://api.robinhood.com/options/events/',
            callback=option_event_from_data,
            result_key='results',
        )


def account_obj_from_data(data: Any) -> AccountObj:
    return AccountObj(
        buying_power=float(data['buying_power']),
        crypto_buying_power=float(data['crypto_buying_power']),
        margin_balance=margin_balance_from_data(data=data['margin_balances']),
        cash=float(data['cash']),
        cash_held_for_orders=float(data['cash_held_for_orders']),
        cash_available_for_withdrawal=float(data['cash_available_for_withdrawal']),
        portfolio_url=data['portfolio'],
        positions_url=data['positions'],
        url=data['url'],
        user_url=data['user'],
        state=data['state'],
        account_type=data['type'],
        option_level=data['option_level'],
        locked=data['locked'],
        deactivated=data['deactivated'],
        permanently_deactivated=data['permanently_deactivated'],
        rhs_account_number=int(data['rhs_account_number']),
        account_number=data['account_number'],
        active_subscription_id=data['active_subscription_id'],
        withdrawal_halted=data['withdrawal_halted'],
        sma=int(data['sma']),
        sweep_enabled=data['sweep_enabled'],
        deposit_halted=data['deposit_halted'],
        max_ach_early_access_amount=float(data['max_ach_early_access_amount']),
        only_position_closing_trades=data['only_position_closing_trades'],
        sma_held_for_orders=int(data['sma_held_for_orders']),
        unsettled_debit=float(data['unsettled_debit']),
        is_pinnacle_account=data['is_pinnacle_account'],
        uncleared_deposits=float(data['uncleared_deposits']),
        unsettled_funds=float(data['unsettled_funds']),
        created_at=parse_datetime_with_microsecond(data['created_at']),
        updated_at=parse_datetime_with_microsecond(data['updated_at']),
    )



import json
import os
from datetime import date
from datetime import datetime
from typing import Any
from typing import Dict
from typing import Optional

from data.redis import redis
from robinhood.lib.errors import URLNotFoundError
from robinhood.lib.helpers import parse_date
from robinhood.lib.helpers import parse_datetime_with_microsecond
from robinhood.lib.marketplace.minticks import min_ticks_from_data
from robinhood.lib.marketplace.minticks import MinTicks
from robinhood.lib.marketplace.optionmarketdata import option_market_data_from_data
from robinhood.lib.marketplace.optionmarketdata import OptionMarketData
from robinhood.lib.robinhoodclient import rhc


class OptionInstrument(object):

    def __init__(
        self,
        instrument_id: str,
        instrument_type: str,
        issue_date: date,
        state: str,
        tradability: str,
        rhs_tradability: str,
        chain_symbol: str,
        exp_date: str,
        strike_price: float,
        chain_id: str,
        url: str,
        min_ticks: MinTicks,
        created_at: datetime,
        updated_at: datetime,
    ) -> None:
        self._instrument_id = instrument_id
        self._instrument_type = instrument_type
        self._issue_date = issue_date
        self._state = state
        self._tradability = tradability
        self._rhs_tradability = rhs_tradability
        self._chain_symbol = chain_symbol
        self._exp_date = exp_date
        self._strike_price = strike_price
        self._chain_id = chain_id
        self._url = url
        self._min_ticks = min_ticks
        self._created_at = created_at
        self._updated_at = updated_at
        self._url = url

    @property
    def instrument_id(self) -> str:
        return self._instrument_id

    @property
    def instrument_type(self) -> str:
        return self._instrument_type

    @property
    def instrument_type_short(self) -> str:
        return self._instrument_type.upper()[0]

    @property
    def instrument_type_full(self) -> str:
        return self._instrument_type.upper()

    @property
    def issue_date(self) -> date:
        return self._issue_date

    @property
    def state(self) -> str:
        return self._state

    @property
    def tradability(self) -> str:
        return self._tradability

    @property
    def rhs_tradability(self) -> str:
        return self._rhs_tradability

    @property
    def chain_symbol(self) -> str:
        return self._chain_symbol

    @property
    def chain_id(self) -> str:
        return self._chain_id

    @property
    def exp_date(self) -> str:
        return self._exp_date

    @property
    def days_to_expire(self) -> int:
        exp_date = datetime.strptime(self.exp_date, '%Y-%m-%d')
        diff = exp_date - datetime.now()
        return diff.days + 1

    @property
    def formatted_exp_date(self) -> str:
        return self._exp_date.replace('-', '')

    @property
    def strike_price(self) -> float:
        return self._strike_price

    @property
    def url(self) -> str:
        return self._url

    @property
    def min_ticks(self) -> MinTicks:
        return self._min_ticks

    @property
    def created_at(self) -> datetime:
        return self._created_at

    @property
    def updated_at(self) -> datetime:
        return self._updated_at

    @property
    def symbol(self) -> str:
        return (
            f'{self._chain_symbol}{self.formatted_exp_date}'
            f'{self.instrument_type_short}{self._strike_price}'
        )

    @property
    def market_data(self) -> Optional[OptionMarketData]:
        return self._get_market_data()

    def _get_market_data(self) -> Optional[OptionMarketData]:
        try:
            market_data: OptionMarketData = rhc.get_object(
                url=self._get_market_data_endpoint(),
                callback=option_market_data_from_data,
            )
            return market_data
        except URLNotFoundError:
            return None

    def _get_market_data_endpoint(self) -> str:
        return f'https://api.robinhood.com/marketdata/options/{self._instrument_id}/'


def option_instrument_from_data(data: Dict[str, Any]) -> OptionInstrument:
    min_ticks = min_ticks_from_data(data=data['min_ticks'])
    return OptionInstrument(
        instrument_id=data['id'],
        instrument_type=data['type'],
        issue_date=parse_date(data['issue_date']),
        state=data['state'],
        tradability=data['tradability'],
        rhs_tradability=data['rhs_tradability'],
        chain_symbol=data['chain_symbol'],
        exp_date=data['expiration_date'],
        strike_price=float(data['strike_price']),
        chain_id=data['chain_id'],
        url=data['url'],
        min_ticks=min_ticks,
        created_at=parse_datetime_with_microsecond(data['created_at']),
        updated_at=parse_datetime_with_microsecond(data['updated_at']),
    )


def option_instrument_from_url(url: str) -> OptionInstrument:
    option_instrument_id = os.path.basename(os.path.normpath(url))
    redis_key = f'ROBINHOOD:OPTION_INSTRUMENT:{option_instrument_id}'
    redis_cache = redis.get(redis_key)
    if redis_cache is None:
        raw = rhc.get_url(url, return_raw_response=True)
        redis.set(name=redis_key, value=raw)
    else:
        raw = redis_cache

    data = json.loads(raw.decode('utf-8'))
    return option_instrument_from_data(data=data)



from report.accountsummary.accountsummary import AccountSummary # noqa
from report.optionsellscreener.optionsellscreener import OptionSellScreener # noqa
from report.symbolhistory.symbolhistory import SymbolHistory # noqa
from robinhood.lib.account.account import account as robinhood_account # noqa



import os
from datetime import datetime
from typing import Any
from typing import Optional

from robinhood.lib.helpers import parse_optional_float


class Equity(object):

    def __init__(
        self,
        symbol: str,
        instrument_id: str,
        ask_price: float,
        ask_size: int,
        bid_price: float,
        bid_size: int,
        last_trade_price: float,
        last_extended_hours_trade_price: Optional[float],
        previous_close: float,
        adjusted_previous_close: float,
        previous_close_date: datetime,
        trading_halted: bool,
        has_traded: bool,
        last_trade_price_source: str,
        updated_at: datetime,
    ) -> None:
        self._symbol = symbol
        self._instrument_id = instrument_id
        self._ask_price = ask_price
        self._ask_size = ask_size
        self._bid_price = bid_price
        self._bid_size = bid_size
        self._last_trade_price = last_trade_price
        self._last_extended_hours_trade_price = last_extended_hours_trade_price
        self._previous_close = previous_close
        self._adjusted_previous_close = adjusted_previous_close
        self._previous_close_date = previous_close_date
        self._trading_halted = trading_halted
        self._has_traded = has_traded
        self._last_trade_price_source = last_trade_price_source
        self._updated_at = updated_at

    @property
    def symbol(self) -> str:
        return self._symbol

    @property
    def instrument_id(self) -> str:
        return self._instrument_id

    @property
    def last_trade_price(self) -> float:
        return self._last_trade_price

    @property
    def last_price(self) -> float:
        if self._last_extended_hours_trade_price is None:
            return self._last_trade_price
        else:
            return self._last_extended_hours_trade_price

    @property
    def dod_percentage(self) -> float:
        return self.last_price / self._adjusted_previous_close - 1


def equity_from_data(data: Any) -> Equity:
    return Equity(
        symbol=data['symbol'],
        instrument_id=os.path.basename(os.path.normpath(data['instrument'])),
        ask_price=float(data['ask_price']),
        ask_size=data['ask_size'],
        bid_price=float(data['bid_price']),
        bid_size=data['bid_size'],
        last_trade_price=float(data['last_trade_price']),
        last_extended_hours_trade_price=parse_optional_float(data['last_extended_hours_trade_price']),
        previous_close=float(data['previous_close']),
        adjusted_previous_close=float(data['adjusted_previous_close']),
        previous_close_date=datetime.strptime(data['previous_close_date'], '%Y-%m-%d'),
        trading_halted=data['trading_halted'],
        has_traded=data['has_traded'],
        last_trade_price_source=data['last_trade_price_source'],
        updated_at=datetime.strptime(data['updated_at'], '%Y-%m-%dT%H:%M:%SZ'),
    )



from report.accountsummary.accountsummary import AccountSummary
from report.symbolhistory.symbolhistory import SymbolHistory
from robinhood.lib.account.account import account


account.update_in_parallel()
symbol_history = SymbolHistory(
    stock_orders=account.stock_orders,
    option_orders=account.option_orders,
    option_events=account.option_events,
)
report = AccountSummary(symbol_history=symbol_history)


def test_gen_reports() -> None:
    text = report._gen_reports()
    for s in {
        'Stock Value',
        'Option Value',
        'Margin Used',
        'Account Value',
        '# of Symbols',
        'Profit on Stocks',
        'Profit on Options',
        'Profit on Call Writes',
        'Profit on Put Writes',
        'Symbol',
        '# of Shares',
        'Settled Profit',
        'Last Price',
        'Buying Power',
    }:
        assert s in text



from alphavantage.alphavantageclient import avc


def test_get_price_history() -> None:
    price_history = avc.get_price_history(
        symbol='ADBE',
        time_window='daily',
        output_size='compact',
    )
    assert price_history.metadata.information == 'Daily Time Series with Splits and Dividend Events'
    assert len(price_history.time_series) > 0



from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from robinhood.lib.account.optionorder.optionleg import option_leg_from_data
from robinhood.lib.account.optionorder.optionleg import OptionLeg
from robinhood.lib.account.optionorder.optionrollingorder import derive_option_rolling_order
from robinhood.lib.account.optionorder.optionrollingorder import OptionRollingOrder
from robinhood.lib.helpers import parse_datetime_with_microsecond
from robinhood.lib.helpers import parse_optional_float


class OptionOrder(object):

    def __init__(
        self,
        order_id: str,
        order_type: str,
        ref_id: str,
        state: str,
        trigger: str,
        chain_symbol: str,
        chain_id: str,
        opening_strategy: Optional[str],
        closing_strategy: Optional[str],
        direction: str,
        response_category: str,
        time_in_force: str,
        premium: Optional[float],
        processed_premium: float,
        price: Optional[float],
        stop_price: Optional[float],
        quantity: float,
        pending_quantity: float,
        processed_quantity: float,
        canceled_quantity: float,
        legs: List[OptionLeg],
        created_at: datetime,
        updated_at: datetime,
    ) -> None:
        self._order_id = order_id
        self._order_type = order_type
        self._ref_id = ref_id
        self._state = state
        self._trigger = trigger
        self._chain_symbol = chain_symbol
        self._chain_id = chain_id
        self._opening_strategy = opening_strategy
        self._closing_strategy = closing_strategy
        self._direction = direction
        self._response_category = response_category
        self._time_in_force = time_in_force
        self._premium = premium
        self._processed_premium = processed_premium
        self._price = price
        self._stop_price = stop_price
        self._quantity = quantity
        self._pending_quantity = pending_quantity
        self._processed_quantity = processed_quantity
        self._canceled_quantity = canceled_quantity
        self._legs = legs
        self._created_at = created_at
        self._updated_at = updated_at

    @property
    def order_id(self) -> str:
        return self._order_id

    @property
    def chain_symbol(self) -> str:
        return self._chain_symbol

    @property
    def state(self) -> str:
        return self._state

    @property
    def direction(self) -> str:
        return self._direction

    @property
    def opening_strategy(self) -> Optional[str]:
        return self._opening_strategy

    @property
    def closing_strategy(self) -> Optional[str]:
        return self._closing_strategy

    @property
    def opening_strategy_name(self) -> Optional[str]:
        if self._opening_strategy is None:
            return None

        return self._opening_strategy.replace('_', ' ').upper()

    @property
    def closing_strategy_name(self) -> Optional[str]:
        if self._closing_strategy is None:
            return None

        return self._closing_strategy.replace('_', ' ').upper()

    @property
    def legs(self) -> List[OptionLeg]:
        return self._legs

    @property
    def opening_legs(self) -> List[OptionLeg]:
        return [
            leg for leg in self.legs
            if leg.position_effect == 'open'
        ]

    @property
    def quantity(self) -> float:
        return self._quantity

    @property
    def processed_quantity(self) -> float:
        return self._processed_quantity

    @property
    def price(self) -> Optional[float]:
        if self._price is None:
            return None

        sign = 1 if self._direction == 'debit' else -1
        return self._price * sign

    @property
    def updated_at(self) -> datetime:
        return self._updated_at

    @property
    def rolling_order(self) -> Optional[OptionRollingOrder]:
        return derive_option_rolling_order(option_order=self)

    @property
    def num_of_legs(self) -> int:
        return len(self._legs)

    @property
    def is_buy_to_close(self) -> bool:
        return (
            self.direction == 'debit'
            and len(self.legs) == 1
            and self.legs[0].side == 'buy'
            and self.legs[0].position_effect == 'close'
        )

    @property
    def is_fig_leaf_open(self) -> bool:
        if not self.is_custom_open_strategy:
            return False

        if self.num_of_legs != 2:
            return False

        buy_leg = None
        sell_leg = None
        for leg in self.legs:
            if not leg.is_open:
                return False

            if leg.is_buy:
                buy_leg = leg

            if leg.is_sell:
                sell_leg = leg

        if buy_leg is None or sell_leg is None:
            return False

        return buy_leg.option_instrument.exp_date > sell_leg.option_instrument.exp_date

    @property
    def is_fig_leaf_close(self) -> bool:
        if not self.is_custom_close_strategy:
            return False

        if self.num_of_legs != 2:
            return False

        buy_leg = None
        sell_leg = None
        for leg in self.legs:
            if not leg.is_close:
                return False

            if leg.is_buy:
                buy_leg = leg

            if leg.is_sell:
                sell_leg = leg

        if buy_leg is None or sell_leg is None:
            return False

        return buy_leg.option_instrument.exp_date < sell_leg.option_instrument.exp_date

    @property
    def is_custom_open_strategy(self) -> bool:
        return self.opening_strategy == 'custom'

    @property
    def is_custom_close_strategy(self) -> bool:
        return self.closing_strategy == 'custom'

    @property
    def fig_leaf_symbol(self) -> str:
        assert self.is_fig_leaf_open or self.is_fig_leaf_close

        buy_leg = None
        sell_leg = None
        for leg in self.legs:
            if leg.is_buy:
                buy_leg = leg

            if leg.is_sell:
                sell_leg = leg

        assert buy_leg is not None
        assert sell_leg is not None
        if self.is_fig_leaf_open:
            near_term_symbol = sell_leg.option_instrument.symbol
            far_term_strike_price = buy_leg.option_instrument.strike_price
            far_term_exp_date = buy_leg.option_instrument.formatted_exp_date
        else:
            near_term_symbol = buy_leg.option_instrument.symbol
            far_term_strike_price = sell_leg.option_instrument.strike_price
            far_term_exp_date = sell_leg.option_instrument.formatted_exp_date

        return self.get_two_leg_symbol(
            left_symbol=near_term_symbol,
            right_exp_date=far_term_exp_date,
            right_instrument_type='C',
            right_strike_price=far_term_strike_price,
        )

    @property
    def spread_symbol(self) -> str:
        if self.num_of_legs != 2:
            raise ValueError('Spread must have two legs.')

        buy_leg = None
        sell_leg = None
        for leg in self.legs:
            if leg.is_buy:
                buy_leg = leg
            else:
                sell_leg = leg
        if buy_leg is None or sell_leg is None:
            raise ValueError('Spread must have one leg buy and one leg sell.')

        buy_strike_type = buy_leg.option_instrument.instrument_type_short
        sell_strike_type = sell_leg.option_instrument.instrument_type_short
        if buy_strike_type != sell_strike_type:
            raise TypeError('Spread must have the same instrument type.')
        instrument_type = buy_strike_type
        buy_strike_price = buy_leg.option_instrument.strike_price
        sell_strike_price = sell_leg.option_instrument.strike_price
        if buy_strike_price == sell_strike_price:
            raise TypeError('Spread must have two legs with different strike price.')

        if instrument_type == 'C' and buy_strike_price < sell_strike_price:
            _left = buy_leg
            _right = sell_leg
        elif instrument_type == 'C' and buy_strike_price > sell_strike_price:
            _left = sell_leg
            _right = buy_leg
        elif instrument_type == 'P' and buy_strike_price < sell_strike_price:
            _left = sell_leg
            _right = buy_leg
        elif instrument_type == 'P' and buy_strike_price > sell_strike_price:
            _left = buy_leg
            _right = sell_leg

        return self.get_two_leg_symbol(
            left_symbol=_left.option_instrument.symbol,
            right_exp_date=_right.option_instrument.formatted_exp_date,
            right_instrument_type=instrument_type,
            right_strike_price=_right.option_instrument.strike_price,
        )

    @property
    def concatenated_symbol(self) -> str:
        components: List[str] = []
        for leg in self.legs:
            direction = '' if leg.is_buy else '-'
            symbol = leg.option_instrument.symbol
            ratio = '' if leg.ratio_quantity == 1 else f'x{leg.ratio_quantity}'
            components.append(f'{direction}{symbol}{ratio}')

        components.sort()
        return '/'.join(components)

    def get_two_leg_symbol(
        self,
        left_symbol: str,
        right_exp_date: str,
        right_instrument_type: str,
        right_strike_price: float,
    ) -> str:
        return (
            f'{left_symbol}/'
            f'{right_exp_date}{right_instrument_type}{right_strike_price}'
        )


def option_order_from_data(data: Dict[str, Any]) -> OptionOrder:
    legs = [
        option_leg_from_data(data=row)
        for row in data['legs']
    ]
    return OptionOrder(
        order_id=data['id'],
        order_type=data['type'],
        ref_id=data['ref_id'],
        state=data['state'],
        trigger=data['trigger'],
        chain_symbol=data['chain_symbol'],
        chain_id=data['chain_id'],
        opening_strategy=data['opening_strategy'],
        closing_strategy=data['closing_strategy'],
        direction=data['direction'],
        response_category=data['response_category'],
        time_in_force=data['time_in_force'],
        premium=parse_optional_float(data['premium']),
        processed_premium=float(data['processed_premium']),
        price=parse_optional_float(data['price']),
        stop_price=parse_optional_float(data['stop_price']),
        quantity=float(data['quantity']),
        pending_quantity=float(data['pending_quantity']),
        processed_quantity=float(data['processed_quantity']),
        canceled_quantity=float(data['canceled_quantity']),
        legs=legs,
        created_at=parse_datetime_with_microsecond(data['created_at']),
        updated_at=parse_datetime_with_microsecond(data['updated_at']),
    )



import os
from datetime import datetime
from typing import Any
from typing import Dict

from data.watchlist import ALL_REVERSED
from robinhood.lib.helpers import parse_datetime_with_microsecond


class StockPosition(object):

    def __init__(
        self,
        instrument_url: str,
        quantity: float,
        position_url: str,
        account_url: str,
        intraday_quantity: float,
        average_buy_price: float,
        pending_average_buy_price: float,
        intraday_average_buy_price: float,
        shares_pending_from_options_events: float,
        shares_held_for_options_collateral: float,
        shares_held_for_options_events: float,
        shares_held_for_stock_grants: float,
        shares_held_for_buys: float,
        shares_held_for_sells: float,
        created_at: datetime,
        updated_at: datetime,
    ) -> None:
        self._instrument_url = instrument_url
        self._quantity = quantity
        self._position_url = position_url
        self._account_url = account_url
        self._intraday_quantity = intraday_quantity
        self._average_buy_price = average_buy_price
        self._pending_average_buy_price = pending_average_buy_price
        self._intraday_average_buy_price = intraday_average_buy_price
        self._shares_pending_from_options_events = shares_pending_from_options_events
        self._shares_held_for_options_collateral = shares_held_for_options_collateral
        self._shares_held_for_options_events = shares_held_for_options_events
        self._shares_held_for_stock_grants = shares_held_for_stock_grants
        self._shares_held_for_buys = shares_held_for_buys
        self._shares_held_for_sells = shares_held_for_sells
        self._created_at = created_at
        self._updated_at = updated_at

    @property
    def instrument_id(self) -> str:
        return os.path.basename(os.path.normpath(self._instrument_url))

    @property
    def symbol(self) -> str:
        return ALL_REVERSED[self.instrument_id]

    @property
    def quantity(self) -> float:
        return self._quantity


def stock_position_from_data(data: Dict[str, Any]) -> StockPosition:
    return StockPosition(
        instrument_url=data['instrument'],
        quantity=float(data['quantity']),
        position_url=data['url'],
        account_url=data['account'],
        intraday_quantity=float(data['intraday_quantity']),
        average_buy_price=float(data['average_buy_price']),
        pending_average_buy_price=float(data['pending_average_buy_price']),
        intraday_average_buy_price=float(data['intraday_average_buy_price']),
        shares_pending_from_options_events=float(data['shares_pending_from_options_events']),
        shares_held_for_options_collateral=float(data['shares_held_for_options_collateral']),
        shares_held_for_options_events=float(data['shares_held_for_options_events']),
        shares_held_for_stock_grants=float(data['shares_held_for_stock_grants']),
        shares_held_for_buys=float(data['shares_held_for_buys']),
        shares_held_for_sells=float(data['shares_held_for_sells']),
        created_at=parse_datetime_with_microsecond(data['created_at']),
        updated_at=parse_datetime_with_microsecond(data['updated_at']),
    )
